#if !defined __BASEPARSER_H__
#define __BASEPARSER_H__

#include <queue>
#include <functional>
#include <numeric>
#include <ext/functional>
#include "iostring.h"
#include "scanner.h"

class SlocString: public LocString {
	bool m_indeed;
	LocString &locString() {return *this;}
public:
	SlocString():LocString("",0,0),m_indeed(false) {}
	SlocString(const SlocString &sls):LocString(sls),m_indeed(sls.m_indeed) {}
	SlocString(bool indeed,const LocString &ls):LocString(ls),m_indeed(indeed) {}
	SlocString &operator=(const SlocString &sls) {
		LocString::operator=(sls);
		m_indeed=sls.m_indeed;
		return *this;
	}

	operator bool() const {return m_indeed;}
	std::ostream &print(std::ostream &os) const;
	const LocString &locString() const {return *this;}
};

inline std::ostream &operator<<(std::ostream &os,const SlocString &sls) {
	return sls.print(os);
}

class QueueScanner {
	Scanner scanner;
	std::queue<Token>	m_queue;
	void clearQueue();
	
public:
	QueueScanner() {}
	QueueScanner(std::istream &is):scanner(is) {}
	void load(std::istream &is) {clearQueue();scanner.load(is);}
	void reset() {clearQueue();scanner.reset();}

	const std::string &inputLine(int k) const {return scanner.inputLine(k);}

	Token getToken();
	void ungetToken(const Token &token);


};

struct ParserBase {

	class Error {
		std::string m_errorMsg;
		int m_line,m_col;
  public:
		Error():m_errorMsg(""),m_line(0),m_col(0) {}
		void set(const std::string &errorMsg,const SlocString &sls) {
			m_errorMsg=errorMsg;m_line=sls.line();m_col=sls.col();
		}
		std::string get() const {
			return m_errorMsg;
		}
		int line() const {return m_line;}
		int col() const {return m_col;}
	};

private:

	QueueScanner &m_scanner;
	Error &m_error;

	SlocString digits();
	const char *validSymb(const LocString &ls,const char *valid);

protected:
	QueueScanner &scanner() {return m_scanner;}
public:
	ParserBase(QueueScanner &scanner, Error &err):m_scanner(scanner),m_error(err) {}

	Token getToken() {return m_scanner.getToken();}
	void ungetToken(const Token &token) {m_scanner.ungetToken(token);}

	const std::string &inputLine(int k) const {return m_scanner.inputLine(k);}

	SlocString symb(const char *valid);
	static char selSymb(const SlocString &);

	SlocString end();
	SlocString name();

	SlocString E();
	SlocString uNumber();
	SlocString iNumber();
	SlocString fNumber();
	SlocString number(bool exact=true) {return exact?iNumber():fNumber();}

	Error &error() {return m_error;}

};

template<class T>
class IntExprParser: public ParserBase {

	bool m_exact;
	T m_zero;
	T m_unit;

	T power(T v,int n) const;
public:
	IntExprParser(  QueueScanner &scanner,
								  Error &err,
									bool exact=true,
									const T &zero = __gnu_cxx::identity_element(plus<T>()),
									const T &unit = __gnu_cxx::identity_element(multiplies<T>())
							 ):ParserBase(scanner,err),m_exact(exact),m_zero(zero),m_unit(unit) {}

	bool expr(T &val) {
		val=zero();
		return sum(val);
	}

	bool atom(T &);
	bool exp(T &);
	bool mul(T &);
	bool sum(T &);

	bool exact() const {return m_exact;}
	T zero() const {return m_zero;}
	T unit() const {return m_unit;}

};


template<class T>
bool IntExprParser<T>::atom(T &val) {
	SlocString n=number(exact());
	if (n) stringScan(n.data(),val);
	else {
		SlocString obrace=symb("(");
		if (!obrace) {
			error().set("( expected",obrace);
			return false;
		}
		bool se=sum(val);
		if (!se) return false;
		SlocString cbrace=symb(")");
		if (!cbrace) {
			error().set(") expected",obrace);
			return false;				
		}
	}
	return true;
}

template<class T>
bool IntExprParser<T>::exp(T &val) {
	bool a=atom(val);
	if (!a) return false;
	SlocString sign=symb("^");
	if (sign) {
		int pow=0;
		IntExprParser<int> powParser( scanner(),error(),true );
		a=powParser.atom(pow);
		if (!a) return false;
		if (pow>0) {
			val=power(val,pow);
		}
		else {
  		if (val==zero()) {
				error().set("the non-positive power of zero is undefined",sign);
				return false;
			}
  		if (pow<0) {
  			val=unit()/val;
  			pow=-pow;
  		}
			else val=unit();
		}
	}
	return true;
}

template<class T>
bool IntExprParser<T>::mul(T &val) {
	bool first=exp(val);
	if (!first) {
		return false;
	}
	for(;;) {
		SlocString sign=symb("*/");
		if (!sign) break;
		T val1=zero();
		bool next=exp(val1);
		if (!next) {
			return false;
		}
		if ( selSymb(sign)=='*' ) val*=val1;
		else {
			if (val1==zero()) {
				error().set("can not divide by zero",sign);
				return false;
			}
			val/=val1;
		}
	}
	return true;
}

template<class T>
bool IntExprParser<T>::sum(T &val) {
	SlocString sign=symb("+-");
	bool first=mul(val);
	if (!first) return false;
	if ( selSymb(sign)=='-' ) val=-val;
	for(;;) {
		sign=symb("+-");
		if (!sign) break;
		T val1=zero();
		bool next=mul(val1);
		if (!next) return false;
		if ( selSymb(sign)=='-' ) val-=val1;else val+=val1;
	}
	return true;
}

template<class T>
T IntExprParser<T>::power(T v, int n) const {
	if (n<0) {
		PRECONDITION(v!=zero());
		v=unit()/v;
		n=-n;
	}
	else
	if (n>0) {

    while ((n&1) == 0) {
      n>>=1;
      v = v*v;
    }
    T result = v;
    n>>=1;
    while (n != 0) {
      v = v*v;
      if ((n & 1) != 0) result = result*v;
      n>>=1;
    }
    return result;
	}
	return unit();
}

template<class T>
class ExprParser: public IntExprParser<T> {

	QueueScanner m_scanner;
  typedef IntExprParser<T> Inherited;
  typedef typename Inherited::Error Error;
	Error m_error;
  
public:
	ExprParser(	bool exact=true,									
							const T &zero = identity_element(plus<T>()),
							const T &unit = identity_element(multiplies<T>())
						):IntExprParser<T>(m_scanner,m_error,exact,zero,unit) {}
	ExprParser(	std::istream &is,
							bool exact=true,
							const T &zero = identity_element(plus<T>()),
							const T &unit = identity_element(multiplies<T>())
						):IntExprParser<T>(m_scanner,m_error,exact,zero,unit),m_scanner(is) {}
	void load(std::istream &is) {m_scanner.load(is);}
	void reset() {m_scanner.reset();}
};

#endif //__BASEPARSER_H__
