#if !defined(__BASIS_H__)
#define __BASIS_H__

#include <ostream>
#include <list>
#include <map>
#include <functional>
#include "polin.h"
#include "setofterms.h"

class Pcounter {  
	char m_model;
	linked_object<unsigned> m_count;  
public:
	Pcounter(char m,unsigned c=1)	:m_model(m),m_count(c) {}
	unsigned checkin() {return (*m_count)++;}
  char model() const {return m_model;}
	unsigned count() const {return *m_count;}
};

class Ptag {  
	char m_model;
	unsigned m_index;  
public:
	Ptag(Pcounter &pc):m_model(pc.model()),m_index(pc.checkin()) {}
	void print(std::ostream &os) const {os<<m_model<<m_index;}
};

inline std::ostream &operator<<(std::ostream &os, const Ptag &pt) {
	pt.print(os);
	return os;
}

struct Tpolin: public Polin, public Ptag {  
	Tpolin(const Polin &p, Pcounter &pc):Polin(p),Ptag(pc) {}
	Tpolin(const Polin &p, const Ptag &pt):Polin(p),Ptag(pt) {}
	Polin &polin() {return *this;}
	const Polin &polin() const {return *this;}
	const Ptag &tag() const {return *this;}
	void print(std::ostream &os) const;
};

inline std::ostream &operator<<(std::ostream &os, const Tpolin &bp) {
	bp.print(os);
	return os;
}

class BasisBase: public std::list<Tpolin> {
	typedef std::list<Tpolin> Inherited;
public:

	typedef Inherited::const_iterator iterator;

	BasisBase() = default;
	template <class InputIterator>
		BasisBase(InputIterator first, InputIterator last):Inherited(first,last) {}
	const_iterator erase(const_iterator pos) {
		return Inherited::erase(pos);
	}
	const_iterator insert(const_iterator pos,const_reference bp) {
		return Inherited::insert(pos,bp);
	}
	bool isStable() const;
	bool isZeroDim() const {
		PRECONDITION(isStable());
		return ::isZeroDim(*this,__gnu_cxx::compose1(std::mem_fun_ref(&Eterm::main),std::mem_fun_ref(&Tpolin::lt)));
		
	}
	SetOfTerms nset() const {
		PRECONDITION(isStable());
		return ::nset(*this,__gnu_cxx::compose1(std::mem_fun_ref(&Eterm::main),std::mem_fun_ref(&Tpolin::lt)));
	}
	void print(std::ostream &os) const;	
};

inline std::ostream &operator<<(std::ostream &os, const BasisBase &b) {
	b.print(os);
	return os;
}

class Basis: public BasisBase {

	typedef BasisBase Inherited;

	struct LtermsLess: public std::binary_function<Eterm,Eterm,bool> {
		LtermsLess():m_order(TermOrder::lex()) {}
		bool isLess(const Eterm &et1, const Eterm &et2) const {
			int e1=et1.edegree();
			int e2=et2.edegree();
			if (e1<e2) return true;
			if (e2<e1) return false;
			return m_order.less(et1.term(),et2.term());
		}
		bool operator()(const Eterm &et1, const Eterm &et2) const {
			return isLess(et1,et2);
		}
	private:
		TermOrder m_order;
	};

	class Lterms: public std::multimap<Eterm,const_iterator,LtermsLess> {
		typedef std::multimap<Eterm,Basis::const_iterator,LtermsLess> Inherited;
		typedef Basis::const_iterator basis_iterator;
	public:
		void insert(basis_iterator bi) {
			Inherited::insert(value_type(bi->lt(),bi));
		}
		void erase(basis_iterator bi) {
			const Eterm &lt=bi->lt();
			iterator first=Inherited::lower_bound(lt);
			iterator last=Inherited::upper_bound(lt);
			for (;first!=last;++first) {
				if (first->second==bi) {
					Inherited::erase(first);
					break;
				}
			}
			PRECONDITION(first!=last);
		}

		class div_iterator {
			typedef Lterms::const_iterator Base;
			const Lterms &m_cont;
			Base m_base;
			Eterm m_et;
			void seekDivisor();
		public:
			div_iterator(const Lterms &cont,const Eterm &et):m_cont(cont),m_base(cont.begin()),m_et(et) {
				seekDivisor();
			}
			div_iterator(const Lterms &cont, Base base, const Eterm &et):m_cont(cont),m_base(base),m_et(et) {}// use with caution!
			bool equal(const div_iterator &di) const {
				PRECONDITION(m_et==di.m_et);
				return m_base==di.m_base;
			}
			bool operator==(const div_iterator &di) const {
				return equal(di);
			}
			bool operator!=(const div_iterator &di) const {
				return !equal(di);
			}
			bool operator==(const basis_iterator &bi) const {
				return m_base->second==bi;
			}
			bool operator!=(const basis_iterator &bi) const {
				return m_base->second!=bi;
			}
			friend inline bool operator==(const basis_iterator &bi,const div_iterator &di) {
				return di==bi;
			}
			friend inline bool operator!=(const basis_iterator &bi,const div_iterator &di) {
				return di!=bi;
			}
			Basis::const_reference operator*() const {
				return *m_base->second;
			}
      Basis::const_pointer operator->() const { return &(operator*()); }


			operator basis_iterator() const {
				return m_base->second;
			}
			div_iterator &operator++() {
				++m_base;
				seekDivisor();
				return *this;
			}
			div_iterator operator++(int) {
				div_iterator temp=*this;
				++m_base;
				seekDivisor();
				return temp;
			}
		};

		div_iterator lower_divisor(const Eterm &et) const {
			return div_iterator(*this,et);
		}
		div_iterator upper_divisor(const Eterm &et) const {
			return div_iterator(*this,end(),et);


		}
	};

	Lterms m_lterms;
public:
	
	typedef Lterms::div_iterator div_iterator;

	div_iterator lower_divisor(const Eterm &et) const {
		return m_lterms.lower_divisor(et);
	}
	div_iterator upper_divisor(const Eterm &et) const {
		return m_lterms.upper_divisor(et);
	}

	void clearDivs() {m_lterms.clear();}
	void reDivs();

	Basis() {}
	Basis(const Basis &b):Inherited(b) {
		reDivs();
	}
	template <class InputIterator>
	Basis(InputIterator first, InputIterator last):Inherited(first,last) {
    reDivs();
	}

	void push_front(const_reference bp) {
		Inherited::push_front(bp);
		m_lterms.insert(begin());
	}

	void push_back(const_reference bp) {
		Inherited::push_back(bp);
		m_lterms.insert(--end());
	}

  void pop_front() {
		m_lterms.erase(begin());
		Inherited::pop_front();
	}
 	void pop_back() {
		m_lterms.erase(--end());
		Inherited::pop_back();
	}

	void swap(Basis &b) {
		Inherited::swap(b);
		m_lterms.swap(b.m_lterms);
	}

  const_iterator insert(const_iterator pos,const_reference bp) {
		pos=Inherited::insert(pos,bp);
		m_lterms.insert(pos);
		return pos;
	}

	template<class Iter>
	void insert(Iter first, Iter last) {
		while (first!=last) push_back(*first++);
	}
	const_iterator erase(const_iterator pos) {
		m_lterms.erase(pos);
		return Inherited::erase(pos);
	}

	void clear() {
		m_lterms.clear();
		Inherited::clear();
	}
	
	SetOfEterms supp() const;

};

#endif // __BASIS_H__
