#if !defined __FPOINT_H__
#define __FPOINT_H__

#include <gmp.h>

#include <cmath>
#include <cstdlib>
#include <stdexcept>
#include <functional>
#include <ext/functional>
#include "lrand48.h"


class Fpoint_error : public std::runtime_error {
  static const char * const prefix;
public:
  Fpoint_error (const std::string& what_arg): std::runtime_error (std::string(prefix)+what_arg) { }
};

class Fpoint {

  mpf_t rep;

  Fpoint(const mpf_t arg) {
		mpf_init_set(rep, arg);
  }
    
  class Precision {  
  
		class OutBuf {
	  	unsigned maxd;
	  	unsigned d;
	  	mutable char *b;
	  
	  	inline unsigned prec2dig(unsigned prec) const {return (unsigned)ceil(double(prec)*log(2.0)/log(10.0));}
	  	void new_prec(unsigned prec) {
				maxd = prec2dig(prec);
				d=maxd;
				b=new char[d+10];
	  	}
		public:
	  	OutBuf(unsigned prec) {
				new_prec(prec);
	  	}
	  
	  	void prec(unsigned prec) {
				delete[] b;
				new_prec(prec);
	  	}
	  	unsigned digits() const {return d;}
	  	void digits(unsigned dig) {
				if (maxd<dig) dig=maxd;
				d=dig;
				delete[] b;
				b=new char[d+10];
	  	}
	  
	  	char *text() const {return b;}
	  	~OutBuf() {delete[] b;}
		};
	
		OutBuf outbuf;
		unsigned long bits;
  	public:
		enum {default_prec = 256};
	
		Precision(unsigned long prec = default_prec) noexcept:outbuf(prec) {
	  	set_prec(prec);
		}

		void set_prec(unsigned long prec = default_prec) {
	  	bits=prec;
	  	::mpf_set_default_prec(bits);
	  	outbuf.prec(prec);
		}
		unsigned long get_prec() const {
	  	return bits;
		}
	
		unsigned digits() const {return outbuf.digits();}
		void digits(unsigned dig) {outbuf.digits(dig);}
		char *text() const {return outbuf.text();}
  };
  
  static Precision default_precision;
  static char *outbuf() {return default_precision.text();}
public:

  static void precision(unsigned long prec) {
		default_precision.set_prec(prec);
  }
  
  static unsigned long precision() {
		return default_precision.get_prec();
  }
  
  static unsigned digits() {
		return default_precision.digits();
  }

  static void digits(unsigned dig) {
		default_precision.digits(dig);
  }
  
  Fpoint() {
		mpf_init(rep);
  }

  Fpoint(const Fpoint& f) {
		mpf_init_set(rep, f.rep);
  }

  explicit Fpoint(long i) {   
		mpf_init_set_si(rep, i);
  }

  explicit Fpoint(int i) {   
		mpf_init_set_si(rep, i);
  }

  explicit Fpoint(double d) {   
		mpf_init_set_d(rep, d);
  }
  
  explicit Fpoint(const char* s) {
    mpf_init(rep);
    try {
      set(s);
	  }
    catch (const Fpoint_error&) {
      mpf_clear(rep);
      throw;
    }
  }

  ~Fpoint() {
		mpf_clear(rep);
  }

  Fpoint& operator=(const Fpoint& f) {
		mpf_set (rep,f.rep);
		return *this;
  }

  Fpoint& operator=(long i) {
		mpf_set_si(rep,i);
	 	return *this;
  }

  Fpoint& operator=(int i) {
		mpf_set_si(rep,i);
		return *this;
  }

  Fpoint& operator=(double d) {
		mpf_set_d(rep, d);
		return *this;
  }

  Fpoint& set(const char *s) noexcept(false) {
		if (mpf_set_str(rep, s, 0) < 0)
		throw Fpoint_error("syntax error in string");
    return *this;
  }

  int sign() const {
		return mpf_sgn(rep);
  }

  bool operator!() const {
		return !mpf_sgn(rep);
  }

  explicit operator void*() const {
		return reinterpret_cast<void*>(mpf_sgn(rep));
  }
  
  std::ostream &print(std::ostream &os) const;
  std::istream &scan(std::istream &is) noexcept(false);
  
  static void write(std::ostream& os, const mpf_t src);
  static void read(std::istream& is, mpf_t dst)  noexcept(false);

	Fpoint operator+(const Fpoint &f) const {
		Fpoint r;
 		mpf_add(r.rep, rep, f.rep);
		return r;
	}

  Fpoint &operator+=(const Fpoint &f) {
    mpf_add(rep, rep, f.rep);
    return *this;  
  }
  
	Fpoint operator-(const Fpoint &f) const {
		Fpoint r;
 		mpf_sub(r.rep, rep, f.rep);
		return r;
	}

  Fpoint &operator-=(const Fpoint &f) {
    mpf_sub(rep, rep, f.rep);
    return *this;  
  }
  
	Fpoint operator-() const {
		Fpoint r;
  	mpf_neg(r.rep, rep);
		return r;
	}

  Fpoint &neg_auto() {
    mpf_neg(rep, rep);
    return *this;
  }

	Fpoint abs() const {
		Fpoint r;
  	mpf_abs(r.rep, rep);
		return r;
	}

	Fpoint &abs_auto() {
  	mpf_abs(rep, rep);
		return *this;
	}

	Fpoint operator*(const Fpoint &f) const {
		Fpoint r;
 		mpf_mul(r.rep, rep, f.rep);
		return r;
	}

  Fpoint &operator*=(const Fpoint &f) {
    mpf_mul(rep, rep, f.rep);
    return *this;
  }
  
	Fpoint operator/(const Fpoint &f) const {
		Fpoint r;
 		mpf_div(r.rep, rep, f.rep);
		return r;
	}

  Fpoint &operator/=(const Fpoint &f) {
		mpf_div(rep, rep, f.rep);
		return *this;
  }
  
  inline Fpoint mod(const Fpoint &f) const;

  static Fpoint zero() {return __gnu_cxx::identity_element(std::plus<Fpoint>());}
  static Fpoint unit() {return __gnu_cxx::identity_element(std::multiplies<Fpoint>());}
	static Fpoint random() {return Fpoint(.5+drand48()/2.);}
  
  bool isZero() const {return *this==zero();}
  bool isUnit() const {return *this==unit();}
  
  int compare(const Fpoint& f) const { return mpf_cmp(rep, f.rep); }
  
  bool isEqual(const Fpoint &f) const {return compare(f)==0;}
  bool isLess(const Fpoint &f) const  {return compare(f)<0;}

  bool operator==(const Fpoint &f) const {
		return isEqual(f);
  }

  bool operator<(const Fpoint &f) const {
		return isLess(f);
  }

  bool isEqual(const Fpoint &f,unsigned long prec) const {return mpf_eq(rep,f.rep,prec);}
  
  static bool exact() {return false;}

  int exp2() const {
		long int exp;
		mpf_get_d_2exp(&exp,rep);
		return exp;
	}

	Fpoint mul2exp(int e)	const {
		Fpoint r;
		if (e>0)
			mpf_mul_2exp(r.rep,rep,e);
		else
			mpf_div_2exp(r.rep,rep,-e);
		return r;			
	}

  Fpoint pow(long k) const;

  Fpoint sqrt() const {
		Fpoint r;
  	mpf_sqrt(r.rep, rep);
		return r;
	}

	Fpoint mul_2exp(long exp) const  {
		Fpoint r;
  	if (exp<0)
			mpf_div_2exp(r.rep,rep,-exp);
  	else
			mpf_mul_2exp(r.rep,rep,exp);
		return r;
	}

  double toDouble() const {
		return mpf_get_d(rep);
  }
  
};

inline std::ostream &operator<<(std::ostream &os,const Fpoint &f) {return f.print(os);}
inline std::istream &operator>>(std::istream &os,Fpoint &f)     {return f.scan(os);}

inline bool operator!=(const Fpoint &x, const Fpoint &y) {
  return !(x ==y);
}

inline bool operator>(const Fpoint &x, const Fpoint &y) {
  return y<x;
}

inline bool operator<=(const Fpoint &x, const Fpoint &y) {
  return !(y < x);
}

inline bool operator>=(const Fpoint &x, const Fpoint &y) {
  return !(x < y);
}


#endif // __FPOINT_H__
