#if !defined __INTEGER_H__
#define __INTEGER_H__

#include <gmp.h>

#include <stdexcept>
#include <functional>
#include <ext/functional>
#include "checks.h"


class Integer;

class Rational;

class Integer_error : public std::runtime_error {
  static const char * const prefix;
public:
  Integer_error (const std::string& what_arg): std::runtime_error (std::string(prefix)+what_arg) {}
};

class Integer {
  mpz_t rep;

  Integer(const mpz_t arg) {
		mpz_init_set(rep, arg);
  }
  
  friend class Rational;
  
public:

  Integer() {
		mpz_init(rep);
  }

  Integer(const Integer& i) {
		mpz_init_set(rep, i.rep);
  }

  explicit Integer(long i) {   
		mpz_init_set_si(rep, i);
  }

  explicit Integer(int i) {   
		mpz_init_set_si(rep, i);
  }

  explicit Integer(double d) {   
		mpz_init_set_d(rep, d);
  }
  
  explicit Integer(const char* s) {
		mpz_init(rep);
		try {
      set(s);
	  }
    catch (const Integer_error&) {
      mpz_clear(rep);
      throw;
    }
  }

  explicit Integer(const Rational& r);
  Integer& operator=(const Rational& r);

  ~Integer() {
		mpz_clear(rep);
  }

  Integer& operator=(const Integer& a) {
		mpz_set (rep, a.rep);
		return *this;
  }

  Integer& operator=(long i) {
		mpz_set_si(rep,i);
	 	return *this;
  }

  Integer& operator=(int i) {
		mpz_set_si(rep, i);
		return *this;
  }

  Integer& operator=(double d) {
		mpz_set_d(rep, d);
		return *this;
  }

  Integer& set(const char *s) noexcept(false) {
		if (mpz_set_str(rep, s, 0) < 0) throw Integer_error("syntax error in string");
    return *this;
  }

  int sign() const {
		return mpz_sgn(rep);
  }

  bool operator!() const {
		return !mpz_sgn(rep);
  }

  explicit operator void*() const {
		return reinterpret_cast<void*>(mpz_sgn(rep));
  }
  
  std::ostream &print(std::ostream &os) const;
  std::istream &scan(std::istream &is) noexcept(false);
  
  static void write(std::ostream& os, const mpz_t src);
  static void read(std::istream& is, mpz_t dst, char delim=0, bool allow_sign=true)  noexcept(false);

  Integer operator+(const Integer &z) const {
	  Integer r;
	 	mpz_add(r.rep, rep, z.rep);
		return r;
	}
  
  Integer &operator+=(const Integer &z) {
    mpz_add(rep, rep, z.rep);
    return *this;  
  }
  
  Integer operator-(const Integer &z) const {
		Integer r;
	 	mpz_sub(r.rep, rep, z.rep);
		return r;
	}
  
  Integer &operator-=(const Integer &z) {
    mpz_sub(rep, rep, z.rep);
    return *this;  
  }
  
  Integer operator-() const {
		Integer r;
  	mpz_neg(r.rep, rep);
		return r;
	}
  
  Integer &neg() {
    mpz_neg(rep, rep);
    return *this;
  }
  
  inline Integer abs() const {
		Integer r;
  	mpz_abs(r.rep, rep);
		return r;
	}

  Integer operator*(const Integer &z) const {
		Integer r;
  	mpz_mul(r.rep, rep, z.rep);
		return r;
	}

  Integer &operator*=(const Integer &z) {
    mpz_mul(rep, rep, z.rep);
    return *this;
  }
  
  Integer operator/(const Integer &z) const {
		Integer r;
  	mpz_tdiv_q(r.rep, rep, z.rep);
		return r;		
	}
  
  Integer &operator/=(const Integer &z) {
		mpz_tdiv_q(rep, rep, z.rep);
		return *this;
  }
  
  Integer operator%(const Integer &z) const {
		Integer r;
  	mpz_tdiv_r(r.rep, rep, z.rep);
		return r;		
	}

  Integer &operator%=(const Integer &z) {
		mpz_tdiv_r(rep, rep, z.rep);
		return *this;
  }

  Integer &inc() {
    mpz_add_ui(rep, rep, 1);
    return *this;
  }  
  
  Integer &dec() {
    mpz_sub_ui(rep, rep, 1);
    return *this;
  }

  Integer &operator++() {
		return inc();
  }

  Integer operator++(int) {
		Integer tmp(*this);
		inc();
		return tmp;
	}

  Integer &operator--() {
		return dec();
  }

  inline Integer operator--(int) {
		Integer tmp(*this);
		dec();
		return tmp;
	}
  
  static Integer zero() {return __gnu_cxx::identity_element(std::plus<Integer>());}
  static Integer unit() {return __gnu_cxx::identity_element(std::multiplies<Integer>());;}
  
  bool isZero() const {return *this==zero();}
  bool isUnit() const {return *this==unit();}
  
  int compare(const Integer& z) const { return mpz_cmp(rep, z.rep); }
  
  bool isEqual(const Integer& z) const {return compare(z)==0;}
  bool isLess(const Integer& z) const  {return compare(z)<0;}

  bool operator==(const Integer &z) const {
		return isEqual(z);
  }

  bool operator<(const Integer &z) const {
		return isLess(z);
  }

  double to_double() const {
		return mpz_get_d(rep);
  }

  static bool exact() {return true;}

  Integer gcd(const Integer &z) const {
		Integer r;
  	mpz_gcd(r.rep, rep, z.rep);
		return r;
	}


  friend void gcd_ext(const Integer &a, const Integer &b, Integer &c, Integer &p, Integer &q) {
  	mpz_gcdext(c.rep, p.rep, q.rep, a.rep, b.rep);
	}

  Integer lcm(const Integer &z) const {
		return (*this*z)/gcd(z);
	}
  
  Integer pow(long k) const {
  	PRECONDITION(k>=0);
		Integer r;
  	mpz_pow_ui(r.rep, rep, k);
		return r;
	}

  Integer sqrt() const {
		Integer r;
  	mpz_sqrt(r.rep, rep);
		return r;
	}
};

inline std::ostream &operator<<(std::ostream &os,const Integer &z) {return z.print(os);}
inline std::istream &operator>>(std::istream &os,Integer &z) {return z.scan(os);}


inline bool operator!=(const Integer &x, const Integer &y) {
  return !(x ==y);
}

inline bool operator>(const Integer &x, const Integer &y) {
  return y<x;
}

inline bool operator<=(const Integer &x, const Integer &y) {
  return !(y < x);
}

inline bool operator>=(const Integer &x, const Integer &y) {
  return !(x < y);
}

#endif // __INTEGER_H__
