#include "gcc_version.h"
#if GCC_VERSION>=3002
#include <ext/numeric>
#else
#define __gnu_cxx std
#endif

#include <numeric>
#include "polin.h"

using namespace std;

template<class BOP,class UOP>
Polin Polin::comb(const Polin &p, BOP bop,UOP uop) const {
	Polin result( size()+p.size() );
	iterator k=result.begin();
	const_iterator i1 = begin();
	const_iterator i2 = p.begin();
	const_iterator e1 = end();
	const_iterator e2 = p.end();
	while(i1!=e1 && i2!=e2) {
		if (i1->eterm()<i2->eterm()) *k++=uop(*i2++);
		else
		if (i2->eterm()<i1->eterm()) *k++=*i1++;
		else {
			*k=bop(*i1,*i2);++i1;++i2;
			if (!k->isZero()) ++k;
		}
	}	
	while (i1!=e1) *k++=*i1++;
	while (i2!=e2) *k++=uop(*i2++);
	result.resize( std::distance(result.begin(),k) );
	result.m_sdeg=std::max(sdegree(),p.sdegree());
	return result;
}

Polin Polin::eliminate(const Polin &red,const Coeff &c,const_iterator notch1,const_iterator notch2) const {
//	const_iterator notch1=monom(et);
//	const_iterator notch2=red.monom(et);
	PRECONDITION(notch1!=end() && notch2!=red.end() && notch1->eterm()==notch2->eterm());
	Polin result(size()+red.size()-2);
	iterator part1=eliminate(begin(),notch1,red.begin(),notch2,result.begin(),c);
	iterator part2=eliminate(notch1+1,end(),notch2+1,red.end(),part1,c);
	result.resize( std::distance(result.begin(),part2) );
	result.m_sdeg=std::max(sdegree(),red.sdegree());
	return result;
}

Polin::iterator Polin::eliminate(const_iterator first1,const_iterator last1,const_iterator first2,const_iterator last2,iterator target,const Coeff &c) {
	while(first1!=last1 && first2!=last2) {
		if (first1->eterm()<first2->eterm()) *target++=*first2++*c;
		else
		if (first2->eterm()<first1->eterm()) *target++=*first1++;
		else {
			*target=*first1+*first2*c;++first1;++first2;
			if (!target->isZero()) ++target;
		}
	}	
	while (first1!=last1) *target++=*first1++;
	while (first2!=last2) *target++=*first2++*c;
	return target;
}

Polin Polin::operator+(const Polin &p) const {
	return comb(p,plus<Monom>(),__gnu_cxx::identity<Monom>());
}

Polin Polin::operator-(const Polin &p) const {
	return comb(p,minus<Monom>(),negate<Monom>());
}

Polin Polin::operator-() const {
	Polin result( *this );
	for (iterator i=result.begin();i!=result.end();++i) i->coeff().neg_auto();
	return result;
}

Polin Polin::operator*(const Monom &m) const {
	Polin result( size() );
	iterator k = pol_shift(begin(),end(),result.begin(),m);
	result.resize( std::distance(result.begin(),k) );
	result.sdegree()=sdegree()+m.degree();	
	return result;
}

Polin Polin::operator*(const Eterm &et) const {
	Polin result( size() );
	iterator k = pol_shift(begin(),end(),result.begin(),et);
	result.resize( std::distance(result.begin(),k) );
	result.sdegree()=sdegree()+et.degree();	
	return result;
}

Polin Polin::operator*(const Coeff &c) const {
	Polin result( size() );
	iterator k = pol_shift(begin(),end(),result.begin(),c);
	result.resize( std::distance(result.begin(),k) );
	result.sdegree()=sdegree();
	return result;
}

template<class M>
Polin::iterator Polin::pol_shift(const_iterator b, const_iterator e, iterator r, const M &m) {
	while(b!=e) {
		*r = *b*m;++b;
		if (!r->isZero()) ++r;
	}
	return r;
}

Polin Polin::operator/(const Coeff &c) const {
	Polin result( size() );
	iterator k = divCoeffs(begin(),end(),result.begin(),c);
	result.resize( std::distance(result.begin(),k) );
	result.sdegree()=sdegree();
	return result;
}

Polin::iterator Polin::divCoeffs(const_iterator b, const_iterator e, iterator r, const Coeff &c) {
	PRECONDITION(!c.isZero());
	while (b!=e) {
		*r = *b/c;
		++r;
		++b;
	}
	return r;
}


Polin &Polin::operator/=(const Coeff &c) {
	iterator k = divCoeffsAuto(begin(),end(),c);
	resize( std::distance(begin(),k) );
	return *this;
}

Polin::iterator Polin::divCoeffsAuto(iterator b, iterator e, const Coeff &c) {
	PRECONDITION(!c.isZero());
	while (b!=e) *b++/=c;
	return b;
}

Polin Polin::operator*(const Polin &p) const {
	Polin result = (size()<=p.size())	? mul_req(begin(),end(),p):mul_req(p.begin(),p.end(),*this);
	result.m_sdeg=sdegree()+p.sdegree();
	return result;
}

Polin Polin::mul_req(const_iterator b, const_iterator e, const Polin &p2) {

	size_type n = std::distance(b,e);
	PRECONDITION(n<=p2.size());
	if ( n==0 ) return Polin();
	else
	if (n==1) return p2*(*b);
	else {
		size_type m = n/2;
		return mul_req(b,b+m,p2)+mul_req(b+m,e,p2);
	}
}

Polin Polin::power(int n) const {
	PRECONDITION(n>=0);
	if (n) {
		return __gnu_cxx::power(*this,n);
	}
	else {
		PRECONDITION(!isZero());
		return lm().unit();		
	}
}

int Polin::totalDegree() const {
	int deg=0;
	for(const_iterator i=begin();i!=end();++i) {
		int di=i->tdegree();
		if (deg<di) deg=di;
	}
	return deg;
}

int Polin::ledeg() const {
	const Term t=lt();
	int ed=0;
	for(const_iterator i=begin();i!=end();++i) {
		if (i->eterm().main()==lt().main() && ed<i->edegree()) ed=i->edegree();
	}
	return ed;
}

int Polin::leord() const {
	const Term t=lt();
	int ed=0;
	for(const_iterator i=begin();i!=end();++i) {
		if (i->eterm().main()==lt().main() && ed>i->edegree()) ed=i->edegree();
	}
	return ed;
}

int Polin::lediff() const {
	return ledeg()-leord();
}

int Polin::eorder() const {
	const_iterator m=min_element(begin(),end(),[](const Monom &m1,const Monom &m2) {return m1.edegree()<m2.edegree();});
	return (m==end() ? 0:m->edegree());
}

void Polin::resetEdegree() {
	for (iterator i=begin();i!=end();++i) i->resetEdegree();
	canonize();
}

Polin &Polin::calibrateEdegree(const SetOfEterms &keep) {
	if (!empty() && begin()->context().eto()!=EtermOrder::to()) {
		int eb=context().eb();
		bool mod=false;
        const_iterator n=begin();
		for(iterator i=begin();i!=end();++i) {
      const_iterator m=i;
      if (m==n) {
        SetOfEterms::const_iterator t=keep.upper_bound(m->eterm());
        if (t==keep.begin())   n=end();
        else {
          --t;
          if (*t==m->eterm()) {++n;continue;}
          n=std::lower_bound(m,const_iterator(end()),*t,[](const Monom &m,const Eterm &key) {return m.eterm()>key;});
        }
      }
      int ceb=m->coeff().edeg(eb);
      if (m->eterm().edegree()<ceb) {
        i->eterm().setEdegree(ceb);
        mod=true;
      }
    }
		if (mod) canonize();
	}
	return *this;
}
  
void Polin::canonize() {
 	iterator k=remove_if(begin(),end(),mem_fun_ref(&Monom::isZero));
 	resize( distance(begin(),k) );
 	if (!empty()) {
 		sort(begin(),end(), [](const Monom &m1,const Monom &m2) {return m1.eterm()>m2.eterm();});
 		k=begin();
 		iterator j=k;
 		while (++j!=end()) {
 			if (k->eterm()==j->eterm()) k->coeff()+=j->coeff();
 			else {
 				if (!k->coeff().isZero()) ++k;
 				if (j!=k) *k=*j;
 			}
 		}
 		if (k!=end() && !k->coeff().isZero()) ++k;
 		resize( distance(begin(),k) );
 	}
}

void Polin::print(std::ostream &os ) const {
	if (empty()) os<<"nil";
	else {
		const_iterator b = begin();
		const_iterator e = end();
		os<<*b;
		while (++b!=e) {
			if (b->coeff().sign()>=0) os<<'+';
			os<<*b;
		}
	}
}

bool Polin::isEqual(const Polin &p) const {
  if (begin()==p.begin()) return true;
	if (size()!=p.size()) return false;
	if (! std::equal(begin(),end(),p.begin(),[](const Monom &m1,const Monom &m2) {return m1.eterm()==m2.eterm();}) ) return false;
	return std::equal(begin(),end(),p.begin(),[](const Monom &m1,const Monom &m2) {return m1.coeff()==m2.coeff();});
}

bool Polin::isLess(const Polin &p) const {
	const_iterator first1=begin();
	const_iterator  last1=end();
	const_iterator first2=p.begin();
	const_iterator  last2=p.end();
  if (first1==first2) return false;
	for (;first1!=last1 && first2!=last2;++first1,++first2) {
		if (first1->eterm()<first2->eterm()) return true;
		if (first2->eterm()<first1->eterm()) return false;
	}
	if (size()<p.size()) return true;
	if (p.size()<size()) return false;
	return std::lexicographical_compare(begin(),end(),p.begin(),p.end(),[](const Monom &m1,const Monom &m2) {return m1.coeff()<m2.coeff();});
}



//TODO: replace the for-loop with std::equal
bool Polin::isAlmostZero() const {
	if (empty()) return true;
	const AlmostEqual &aEqual=begin()->context().almostEqual();	
	if (aEqual.scheme()==AlmostEqual::Exact) return false; // not empty <=> not zero - the representation invariant
	for (const_iterator i=begin();i!=end();++i) {
		if ( !aEqual(i->coeff()) ) return false;
	}
	return true;
}

Polin &Polin::normalize() {
	if (!empty()) {
  	const_iterator first=((const Polin *)this)->begin(); // explicit cast to const *
  	const_iterator last=((const Polin *)this)->end();    // to avoid possibly redundant shared_array::divorce()
  	if ( first->coeff().isType<Rational>() ) {
  		Rational r=first->coeff();
  		while (++first!=last) r=r.n_gcd_d_lcm(first->coeff());
  		if ( !r.isUnit() ) {
  			Coeff cr=r;
  			for (iterator upd=begin();upd!=end();++upd) upd->coeff()/=cr;
  		}
  	}
  	else
  	if ( first->coeff().isType<Zp>() ) {
  		if ( !first->coeff().isUnit() ) {
  			Zp z=first->coeff();
  			iterator upd=begin();
  			upd->coeff()=Zp::unit();
  			while (++upd!=end()) upd->coeff()/=z;
  		}
  	}
  	else {
      PRECONDITION(!lc().exact()); 
      if ( isAlmostZero() ) clear();
      else {
    		int exp2=first->coeff().exp2();
    		while (++first!=last) {
    			int e2=first->coeff().exp2();
    			if (exp2<e2) exp2=e2;
    		}
        if (exp2>0) {
    			exp2=-exp2;
    			for (iterator upd=begin();upd!=end();++upd) upd->coeff()=upd->coeff().mul2exp(exp2);
    		}
  		}
  	}
	}
	return *this;
}

SetOfEterms Polin::supp() const {
	SetOfEterms result;
	for (const_iterator i=begin();i!=end();++i) result.insert(i->eterm());
	return result;
}
