#if !defined(__POLIN_H__)
#define __POLIN_H__

#include <iostream>
#include <functional>
#include "shared_object.h"
#include "setofeterms.h"
#include "monom.h"

class Polin {
	typedef shared_array<Monom> Rep;

	Rep m_data;
	int m_sdeg;

  Monom &operator[](unsigned int i) {return m_data[i];}

public:

  typedef Rep::iterator iterator;
  typedef Rep::const_iterator const_iterator;
  typedef Rep::reverse_iterator reverse_iterator;
  typedef Rep::const_reverse_iterator const_reverse_iterator;

	const Rep &data() const {return m_data;}

	typedef Rep::size_type size_type;

private:

	iterator begin() {return m_data.begin();}
	iterator end() {return m_data.end();}

	reverse_iterator rbegin() {return m_data.rbegin();}
	reverse_iterator rend() {return m_data.rend();}

	Polin(size_type n):m_data(n),m_sdeg(0) {}
	void resize(size_type n) {m_data.resize(n);}

	int &sdegree() {return m_sdeg;}

public:

	const_iterator begin() const {return m_data.begin();}
	const_iterator end() const {return m_data.end();}

	const_iterator const_begin() const {return begin();}
	const_iterator const_end() const {return end();}


	const_reverse_iterator rbegin() const {return m_data.rbegin();}
	const_reverse_iterator rend() const {return m_data.rend();}

  const Monom &operator[](unsigned i) const {return m_data[i];}
  size_type size() const {return m_data.size();}
	bool empty() const {return m_data.empty();}	
	void clear() {m_data.clear();}

	Polin():m_sdeg(0) {}
	Polin(const Polin &p):m_data(p.m_data),m_sdeg(p.m_sdeg) {}
	Polin(const Monom &m):m_data(1,m),m_sdeg(m.tdegree()) {}
	Polin(const Eterm &et):m_data(1,Monom(et)),m_sdeg(et.tdegree()) {}
	Polin(const Term &t):m_data(1,Monom(t)),m_sdeg(t.degree()) {}
	

	const PolyContext &context() const {return lm().context();}

	Polin &operator=(const Polin &p) {
		m_data = p.m_data;
		m_sdeg=p.sdegree();
		return *this;
	}

	Polin &operator=(const Monom &m) {
		m_data.assign(1,&m);
		m_sdeg=m.tdegree();
		return *this;
	}

	Polin operator+(const Polin &p) const;
	Polin operator-(const Polin &p) const;
	Polin operator-() const;

	Polin operator*(const Monom &m) const;
	Polin operator*(const Eterm &et) const;
	Polin operator*(const Coeff &c) const;
	Polin operator/(const Coeff &c) const;
	Polin operator*(const Polin &p) const;

	Polin &operator*=(const Coeff &c);
	Polin &operator/=(const Coeff &c);

	bool isEqual(const Polin &p) const;
	bool isLess(const Polin &p) const;
	bool isAlmostZero() const;

	bool operator==(const Polin &p) const {
		return isEqual(p);
	}
	bool operator!=(const Polin &p) const {
		return !isEqual(p);
	}
	bool operator<(const Polin &p) const {
		return isLess(p);
	}
	bool operator>(const Polin &p) const {
		return p.isLess(*this);
	}

	Polin power(int n) const;

	bool isUnit() const {return size()==1 && begin()->isUnit();}
	bool isZero() const {return empty();}

	int totalDegree() const;
	int sdegree() const {return m_sdeg;}
	void resetsdegree() {m_sdeg=totalDegree();}
	void resetEdegree();

	const Monom &lm() const {
		PRECONDITION(!empty());
		return *begin();

	}
	const Eterm &lt() const {return lm().eterm();}
	const Coeff &lc() const {return lm().coeff();}

	int ledeg() const;
	int leord() const;
  int lediff() const;
	int eorder() const;

	void print(std::ostream &os) const;
	Polin &normalize();
	Polin &calibrateEdegree(const SetOfEterms &keep = SetOfEterms());

	SetOfEterms supp() const;

	const_iterator monom(const Eterm &et) const {
    const_iterator hint=std::lower_bound(begin(),end(),et,[](const Monom &m,const Eterm &key) {return m.eterm()>key;});
    return hint!=end() && hint->eterm()==et ? hint:end();
  }

	const Monom &monom(unsigned int i) const {return (*this)[i];}

	Polin eliminate(const Polin &red,const Coeff &c,const Eterm &et) const {
    return eliminate(red,c,monom(et),red.monom(et));
  }
	Polin eliminate(const Polin &red,const Coeff &c, int i1,int i2) const {
    return eliminate(red,c,&monom(i1),&red.monom(i2));
  }
	Polin eliminate(const Polin &red,const Coeff &c, const_iterator m1,const_iterator m2) const;

private:

	void canonize();

	static iterator eliminate(const_iterator first1,const_iterator last1,const_iterator first2,const_iterator last2,iterator result,const Coeff &c);

	static Polin mul_req(const_iterator b, const_iterator e, const Polin &p2);
	template<class M>
	static iterator pol_shift(const_iterator b, const_iterator e, iterator r, const M &m);

	static iterator divCoeffs(const_iterator b, const_iterator e, iterator r, const Coeff &c);
	static iterator divCoeffsAuto(iterator b, iterator e, const Coeff &c);

	template<class BOP,class UOP> Polin comb(const Polin &p,BOP bop, UOP uop) const;

};

inline Polin identity_element(std::plus<Polin>) {return Polin();}
inline Polin identity_element(std::multiplies<Polin>) {return Polin(Monom());}

inline std::ostream &operator<<(std::ostream &os, const Polin &p) {
	p.print(os);
	return os;
}

#endif // __POLIN_H__
