#if !defined __SHARED_OBJECT_
#define __SHARED_OBJECT_

#include <assert.h>
#include <memory>
#include <utility>
#include <iterator>
#include <algorithm>
#include <functional>
#include <set>
#include <typeinfo>
#include <iostream>
#include "const_ref_or_val.h"
#include "malloc_alloc.h"

//#pragma warning (disable: 4200 4786 4291)

template <class Object>
class shared_object {

public:
  typedef Object element_type;
  typedef typename FUNC_ARG(element_type) func_arg;

protected:
  struct rep {      
	int refc;
	Object obj;

	rep() : refc(1) { }
    rep(func_arg obj_arg) :refc(1), obj(obj_arg) { }
    rep(const rep& o) : refc(1), obj(o.obj)  { }
    ~rep() {}
    void* operator new(size_t s) { return malloc_alloc::allocate(s); }
    void operator delete(void* p) { malloc_alloc::deallocate(p,sizeof(rep)); }
  };

  struct Cache {
  
   rep *kept[512];
   
   bool store(rep *r);
   rep *find();
   
   rep *dispatch() {
     rep *r = find();
     if (!r) r=new rep();
     return r;
   }

   rep *dispatch(func_arg obj_arg) {
     rep *r = find();
     if (r) r->obj=obj_arg;
     else r = new rep(obj_arg);
     return r;
   }

   rep *dispatch(const rep &o) {
     rep *r=find();
     if (r) r->obj=o.obj;
     else r = new rep(o);
     return r;
   }

   Cache():preserved(0),pos(0),cs(0),ch(0),cd(0),valid(true) {}

   void flush() {
     while (pos) {
       rep *r=kept[--pos];
       delete r;
     }
   }

   ~Cache() {
     std::cout<<__PRETTY_FUNCTION__<<" cache dispatched "<<cd<<" stored "<<cs<<" hits "<<ch<<'\n'
              <<"shared_object<"<<typeid(Object).name()<<"> preserved "<<preserved<<" copy operations\n";
     flush();
     valid=false;
   }

   int preserved;

  protected:
    unsigned pos;
    unsigned cs,ch,cd;
    bool valid;
  };

  static inline Cache &cache() {
    static Cache __shared_object_cache;
    return __shared_object_cache;
  }
  
   rep *body;
   
   void leave() { assert(body->refc>0);if (! --body->refc) cache().store(body); }
   void divorce() {assert(body->refc>1);--body->refc;body=cache().dispatch(*body);--cache().preserved;}

public:

  typedef Object object_type;
  typedef Object* iterator;
  typedef const Object* const_iterator;

  shared_object() : body(cache().dispatch()) { }
  shared_object(const Object& obj_arg) : body(cache().dispatch(obj_arg)) { }
  shared_object(const shared_object& s) : body(s.body) { ++body->refc;++cache().preserved;}
  ~shared_object() { leave(); }

  shared_object& operator= (const shared_object& s) {
    if (this!=&s) {
      s.body->refc++;
      leave();
      body=s.body;
      ++cache().preserved;
    }
    return *this;
  }

  void swap(shared_object& s2) { std::swap(body,s2.body); }

  static void flush_cache() {cache().flush();}

  iterator operator-> () { if (body->refc > 1) divorce(); return &body->obj; }
  Object& operator* ()  { if (body->refc > 1) divorce(); return  body->obj; }


  const_iterator operator-> () const { return &body->obj; }
  const Object& operator* ()  const { return  body->obj; }
};

template <class Object>
bool shared_object<Object>::Cache::store(rep *r) {
  if (valid && pos<512) {
    kept[pos++]=r;
    ++cs;
    return true;
  }
  else {
    delete r;
    return false;
  }
}

template <class Object>
typename shared_object<Object>::rep * shared_object<Object>::Cache::find() {
  assert(valid);
  ++ch;
  if (!pos) return 0;
  rep *r=kept[--pos];
  ++cd;
  assert(!r->refc);
  r->refc=1;
  return r;
}

template <class Object>
class linked_object: public shared_object<Object> {
  typedef shared_object<Object> Inherited;
  typedef typename Inherited::iterator iterator;
  typedef typename Inherited::const_iterator const_iterator;
  const Inherited &const_inherited() {return *this;}
public:

  linked_object() { }
  linked_object(const Object& obj_arg): Inherited(obj_arg) {}
  linked_object(const linked_object& s) : Inherited(s) {}
  ~linked_object() {}

  iterator operator-> () { return const_cast<iterator>(const_inherited().operator->()); }
  Object& operator* ()  { return const_cast<Object &>(const_inherited().operator*()); }


  const_iterator operator-> () const { return Inherited::operator->(); }
  const Object& operator* ()  const { return  Inherited::operator*(); }

};

template <class Object>
class shared_array {
public:

  typedef Object element_type;
  typedef typename FUNC_ARG(element_type) func_arg;
	typedef size_t size_type;

protected:

  struct rep {
    int refc;
    size_t draft;
    size_t size;
    Object obj[0];

	//Object *begin() {return obj;}
	//Object *end() {return obj+size;}

    rep() : refc(1),draft(0),size(0) { }

    rep(size_t n) : refc(1),draft(n), size(n) {  
      std::uninitialized_fill_n(obj,n,Object());
    }

    /// Creates an array with n elements initialized with the given value.
    rep(size_t n, func_arg init) : refc(1),draft(n), size(n) {  
	  	std::uninitialized_fill_n(obj,n,init);
    }

    /// Creates an array with n elements initialized from an input sequence.
    template <class ForwardIterator>
    rep(size_t n, ForwardIterator src) : refc(1),draft(n),size(n) {
	  	std::uninitialized_copy_n(src, n, obj);
		}

    rep(const rep& o) : refc(1),draft(o.size), size(o.size) {  
	  	std::uninitialized_copy_n(o.obj, o.size, obj);
		}

    static void destroy(Object * a,Object * b) {
      while (a!=b) a++->~Object();
    }

    ~rep() { if (draft) destroy(obj, obj+draft); }

    static size_t total_size(size_t n) {return sizeof(rep)+n*sizeof(Object);}

    void* operator new(size_t, size_t n) { return malloc_alloc::allocate(total_size(n)); }
    void operator delete(void* p) {  malloc_alloc::deallocate((rep *)p,total_size(reinterpret_cast<rep*>(p)->size));}
    void operator delete(void* p,size_t) {  operator delete(p);}
  };

  struct Cache {

    struct rless: public std::less<rep *> {
      typedef std::less<rep *> Inherited;
      typename Inherited::result_type operator()(const typename Inherited::first_argument_type r1,const typename Inherited::second_argument_type r2) const {return r1->draft<r2->draft;}
    };

    std::multiset<rep *,rless> kept;
    typedef typename std::multiset<rep *,rless>::iterator kept_iterator;

    bool store(rep *r);
    rep *find(int n = 0);

    rep *dispatch() {
      rep *r = find();
      if (!r) r=new(0) rep();
      return r;
    }
    
    rep *dispatch(size_t n) {
      rep *r = find(n);
      if (!r) r = new(n) rep(n);
      return r;
    }

    rep *dispatch(size_t n, func_arg init) {
      rep *r = find(n);
      if (r) std::fill_n(r->obj,n,init);
      else r = new(n) rep(n,init);
      return r;
    }

    template<class ForwardIterator>
    rep *dispatch(size_t n,ForwardIterator src,size_t m);

    template<class ForwardIterator>
    rep *dispatch(size_t n,ForwardIterator src);

    rep *dispatch(const rep &o);

    Cache():preserved(0),cs(0),ch(0),cd(0),valid(true) {}

    void flush() {
      while (!kept.empty()) {
        kept_iterator i=kept.begin();
        rep *r = *i;
        kept.erase(i);
        delete r;
      }
    }

    ~Cache() {
      std::cout<<__PRETTY_FUNCTION__<<" cache dispatched "<<cd<<" stored "<<cs<<" hits "<<ch<<'\n'
               <<"shared_array<"<<typeid(Object).name()<<"> preserved "<<preserved<<" copy operations\n";
      flush();
      valid=false;
    }

    int preserved;

  protected:
    unsigned cs,ch,cd;
    bool valid;
  };

  static inline Cache &cache() {
    static Cache __shared_array_cache;
    return __shared_array_cache;
  }

  rep *body;

  static rep *snap(rep *b) {
    assert(b->refc > 0);
    b->refc++;
    ++cache().preserved;
    return b;
  }

  void leave() { 
    assert(body->refc>0);
    if (! --body->refc) cache().store(body); 
  }

  void divorce() {
    assert(body->refc>1);
    body->refc--;
    body=cache().dispatch(*body);
    --cache().preserved;
  }

public:

  typedef Object* iterator;
  typedef const Object* const_iterator;
  typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
  typedef std::reverse_iterator<iterator> reverse_iterator;


  const_iterator begin() const {return **this;}
  const_iterator end() const {return begin()+body->size;}

  iterator begin() {return **this;}
  iterator end() {return begin()+body->size;}

  reverse_iterator rbegin() { return reverse_iterator(end()); }
  const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }

  reverse_iterator rend() { return reverse_iterator(begin()); }
  const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }


  shared_array() : body(cache().dispatch()) { }
  explicit shared_array(size_t n) : body(cache().dispatch(n)) { }

  shared_array(size_t n, func_arg init) : body(cache().dispatch(n,init)) { }

  template <class ForwardIterator>
  shared_array(size_t n, ForwardIterator src) : body(cache().dispatch(n,src)) { }

  shared_array(const shared_array& s) : body(snap(s.body)) { }

  size_t size() const { return body->size; }
  bool empty() const {return size()==0;}
  
  ~shared_array() { leave(); }

  shared_array& operator= (const shared_array& s) {
    if (this!=&s) {
      snap(s.body);
      leave();
      body=s.body;
    }
    return *this;
  }

  template <class ForwardIterator>
  void assign(size_t n, ForwardIterator src) {
    if (body->refc > 1  ||  body->draft < n) {
      leave();
      body=cache().dispatch(n,src);
    } 
    else  {
	  	std::copy_n(src,n,body->obj);
	  	body->size=n;
		}
  }

  static unsigned min(unsigned a,unsigned b) {
    return (a<b) ? a:b;
  }

  void resize(unsigned n) {
		if (body->size==n) return;
    assert(body->size<=body->draft);
    if ((body->refc > 1) || (body->draft<n)) {
     	rep *temp = cache().dispatch(body->size,body->obj,n);
     	leave();
     	body=temp;
    }
    else {
	  	if (body->size<n) std::fill_n(&body->obj[body->size],n-body->size,Object());
     	body->size=n;
		}
  }
	void clear() {resize(0);}

  Object* operator*() { if (body->refc > 1) divorce(); return  body->obj; }
  const Object* operator*() const {return  body->obj; }

  Object &operator[](unsigned int i) {assert(i<body->size);return (**this)[i];}
  const Object &operator[](unsigned i) const {assert(i<body->size);return (**this)[i];}

  void swap(shared_array& a2) { std::swap(this->body,a2.body); }

  static void flush_cache() {cache().flush();}

  void fill(func_arg val) {std::fill(begin(),end(),val);}
  
  friend std::ostream &operator<<(std::ostream &os,const shared_array<Object> &sh) {
    std::copy(sh.begin(),sh.end(),std::ostream_iterator<Object>(os," "));
    return os;
  }
};

template <class Object>
bool shared_array<Object>::Cache::store(rep *r) {
  if (valid && kept.size()<512) {
    kept.insert(r);
    ++cs;
    return true;
  }
  else {
    delete r;
    return false;
  }
}

template <class Object>
typename shared_array<Object>::rep *shared_array<Object>::Cache::find(int n) {
  assert(valid);
  ++ch;
  rep key;key.draft=n;
  kept_iterator p=kept.lower_bound(&key);
  key.draft=0;
  if (p==kept.end()) return 0;
  ++cd;
  rep *r=*p;
  kept.erase(p);
  r->size=n;
  assert(r->size<=r->draft);
  assert(!r->refc);
  r->refc=1;
  return r;
}

template <class Object>
template<class ForwardIterator>

typename shared_array<Object>::rep *shared_array<Object>::Cache::dispatch(size_t n,ForwardIterator src,size_t m) {
  rep *r=find(m);
  if (r) {
    if (n<m) 
      std::fill(std::copy_n(src,n,r->obj),r->obj+m,Object());
    else
      std::copy_n(src,m,r->obj);
  }
  else {
    r = new(m) rep(min(n,m),src);
    if (n<m) std::uninitialized_fill_n(&r->obj[n],m-n,Object());
    r->size=r->draft=m;
  }
  return r;
}

template <class Object>
template<class ForwardIterator>
typename shared_array<Object>::rep *shared_array<Object>::Cache::dispatch(size_t n,ForwardIterator src) {
  rep *r=find(n);
  if (r) 
		std::copy_n(src,n,r->obj);
  else
		r = new(n) rep(n,src);
  return r;
}

template <class Object>
typename shared_array<Object>::rep *shared_array<Object>::Cache::dispatch(const rep &o) {
  rep *r=find(o.size);
  if (r) 
		std::copy_n(o.obj,o.size,r->obj);
  else 
		r = new(o.size) rep(o);
  return r;
}

#endif //__SHARED_OBJECT_
