#include <list>
#include <algorithm>
#include "symbred.h"

using namespace std;

struct BetterReductor: public Report {
	BetterReductor(const Report &r):Report(r) {}
	bool operator()(Basis::const_iterator cur, Basis::const_iterator prev);
};

bool BetterReductor::operator()(Basis::const_iterator cur, Basis::const_iterator prev) {
	if (cur->size()==1 && prev->size()> 1) {
		REPORT(3,cur->tag()<<" is a better reducer because it is a monomial : "<<*cur<<"\n");
		return true;
	}
	if (cur->size()> 1 && prev->size()==1) return false;

	if (prev->lc().abs()<cur->lc().abs()) {
		REPORT(3,cur->tag()<<" is a better reducer because it has a bigger absolute value leading coefficient : "<<*cur<<"\n");
		return true;
	}

  if (cur->lt().edegree()<prev->lt().edegree()) {
		REPORT(3,cur->tag()<<" is a better reducer because it has a smaller e-degree of the leading term : "<<*cur<<"\n");
    return false;
  }
  if (prev->lt().edegree()<cur->lt().edegree()) return false;
	if (cur->lc().abs()<prev->lc().abs()) return false;

	if (prev->size()>1 && cur->size()>1) {
//		Coeff pdc = (prev->lc()-(*prev)[1].coeff()).abs();
//		Coeff cdc = (cur->lc()-(*cur)[1].coeff()).abs();
    Coeff pdc = ((*prev)[1].coeff()/prev->lc()).abs();
    Coeff cdc = ((*cur)[1].coeff()/cur->lc()).abs();
		if (cdc<pdc) {
			REPORT(3,cur->tag()<<" is a better reducer because it has smaller deviation of coefficients : "<<*cur<<"\n");
			return true;
		}
		if (pdc<cdc) return false;
	}

	if (cur->context().eto()==EtermOrder::te()) {
		int led1=cur->lediff();
		int led2=prev->lediff();
		if (led1<led2) {
			REPORT(3,cur->tag()<<" is a better reducer because it has a lower leading e-difference : "<<*cur<<"\n");
			return true;
		}
		if (led2<led1) return false;
	}


	if (prev->lt().tdegree()<cur->lt().tdegree()) {
		REPORT(3,cur->tag()<<" is a better reducer because it has a higher degree of the main part of the leading term : "<<*cur<<"\n");
		return true;
	}
	if (cur->lt().tdegree()<prev->lt().tdegree()) return false;


	if (cur->sdegree()<prev->sdegree()) {
		REPORT(3,cur->tag()<<" is a better reducer because it has a lower sugar degree : "<<*cur<<"\n");
		return true;
	}
	if (prev->sdegree()<cur->sdegree()) return false;


	if (cur->size()<prev->size()) {
		REPORT(3,cur->tag()<<" is a better reducer because it contains less terms : "<<*cur<<"\n");
		return true;
	}
	if (prev->size()<cur->size()) return false;


	return false;
}

struct FindReductor: public Report {

	typedef Basis::div_iterator div_iterator;
	typedef Basis::const_iterator basis_iterator;
	basis_iterator operator()(const Eterm &et,const Basis &basis);

	FindReductor(const Report &r):Report(r),m_better(r) {}

protected:
	BetterReductor m_better;
};

FindReductor::basis_iterator FindReductor::operator()(const Eterm &et,const Basis &basis) {
	basis_iterator result=basis.end();
	div_iterator first=basis.lower_divisor(et);
	div_iterator last=basis.upper_divisor(et);
	for(;first!=last;++first) {
		if (result==basis.end() || m_better(first,result)) {
			REPORT(3 && result==basis.end(),"found a reducer "<<first->tag()<<" for "<<et<<"\nwith "<<*first<<"\n");
			result=first;
		}
	}
	return result;
}

struct AutoFindReductor: public FindReductor {

	AutoFindReductor(const Report &r):FindReductor(r) {}

	basis_iterator operator()(const Eterm &et,
														const Basis &basis,
														basis_iterator self,
														bool top);
};

AutoFindReductor::basis_iterator AutoFindReductor::operator()(const Eterm &et,
																															const Basis &basis,
																															basis_iterator self,
																															bool top) {
	basis_iterator result=basis.end();
	div_iterator first=basis.lower_divisor(et);
	div_iterator last=basis.upper_divisor(et);
	for(;first!=last;++first) {
		if (top && first==self) continue;
		if (result==basis.end() || m_better(first,result)) {
			REPORT(3 && result==basis.end(),"found an auto-reducer "<<first->tag()<<" for "<<et<<"\nwith "<<*first<<"\n");
			result=first;
		}
	}
	return result;
}

int SymbRed::symbRed(Basis &pols, const Basis &basis, const Basis &simp,bool self) {
	REPORT(3,"The set of polynomials for symbolic reduction is\n"<<pols);
	int hits=0;
	FindReductor fr(report());
	AutoFindReductor afr(report());
	BetterReductor better(report());
	for(basis_iterator p=pols.begin();p!=pols.end();++p) {
		Polin::const_iterator first=p->begin();
		Polin::const_iterator last=p->end();
		REPORT(3,"reducing supp("<<p->tag()<<")="<<p->supp()<<"\nwith "<<*p<<"\n" );
		for(;first!=last;++first) {
			const Eterm &et=first->eterm();
			if ( m_done.contains(et) ) continue;
			m_done.insert(et);
			//REPORT(3,"looking for a reducer of "<<et<<" in the basis\n");
			basis_iterator reductor=fr(et,basis);
			bool haveReductor= (reductor!=basis.end());
			if (haveReductor) {
				//REPORT(3,"Found the reducer of "<<et<<" in the basis "<<reductor->tag()<<'\n');
				//REPORT(4,"with "<<*reductor<<'\n');
				if (!simp.empty()) {
					REPORT(3,"looking for a (possibly better) reducer of "<<et<<" in the simplifiers\n");
					basis_iterator simp_reductor=fr(et,simp);
					if ( simp_reductor!=simp.end() && better(simp_reductor,reductor) ) reductor=simp_reductor;
				}
				m_red.insert(et);
				REPORT(3,et<<" is added to the set of reducible terms\n");
			}
			else {
				//REPORT(3,"no reducers found\n");
			}
			bool toKeep=false;
			if (self) {
				//REPORT(3,"looking for a "<<(haveReductor ? "(possibly better)":"")<<" auto-reducer of "<<et<<'\n');
				REPORT(3 && haveReductor,"looking for a (possibly better) auto-reducer of "<<et<<'\n');
				basis_iterator auto_reductor=afr(et,pols,p,first==p->begin());
				if ( auto_reductor!=pols.end() ) {
					//REPORT(3,"Found the auto-reducer "<<auto_reductor->tag()<<'\n');
					//REPORT(4,"with "<<*auto_reductor<<'\n');
					if (auto_reductor->lt()==et) {
						if (!haveReductor) {
							m_keep.insert(et);
							toKeep=true;
							REPORT(3,et<<" is added to the set of terms to keep\n");
						}
					}
					else {
						if (haveReductor) {
							if ( better(auto_reductor,reductor) ) reductor=auto_reductor;
						}
						else {
							reductor=auto_reductor;
							m_red.insert(et);
							REPORT(3,et<<" is added to the set of reducible terms\n");
							haveReductor=true;
						}
					}
				}
				else {
					//REPORT(3,"no auto-reducers found\n");
				}
			}
			if (haveReductor) {
				Eterm s=et.divide(reductor->lt());
				Tpolin shift=s.isUnit() ? *reductor:Tpolin(*reductor*s,counter);
				PRECONDITION(!shift.empty());
				pols.push_back( shift );
				REPORT(3,"Added the reducer for "<<et<<" : "<<s<<" * "<<reductor->tag()<<" = "<<shift<<'\n');
				++hits;
			}
			else
			if (!toKeep && first==p->begin()) m_done.erase(et);
		}
	}
	REPORT(2,"The symbolically reduced set of polynomials is\n"<<pols);
	REPORT(2,"The set of reducible terms is "<<m_red<<'\n');
	REPORT(2,"The set of terms to keep is "<<m_keep<<'\n');
	REPORT(1,"Symbolic reduction finished with "<<hits<<" reducers added to the basis\n");
	stat()["sr"]+=hits;
	return hits;
}

int SymbRed::operator()(Basis &pols, const Basis &basis, const Basis &simp) {
	clear();
	REPORT(1,"Entering symbolic reduction.\n");
	return symbRed(pols,basis,simp,true);
}

int SymbRed::spairs(Basis &pols, const Basis &basis, const Basis &simp) {
	clear();
	Basis::const_iterator first=pols.begin();
	Basis::const_iterator  last=pols.end();
	REPORT(1,"Entering symbolic reduction of s-pairs.\n");
	for (;first!=last;++first) m_red.insert(first->lt());
	m_done=m_red;
	REPORT(3,"The set of leading terms of s-pairs is: "<<m_red<<'\n');
	return symbRed(pols,basis,simp,true);
}

int SymbRed::mutual(Basis &pols) {
	clear();
	REPORT(1,"Entering symbolic auto-reduction.\n");
	return symbRed(pols,Basis(),Basis(),true);
}

void SymbRed::clear() {
	m_done.clear();
	m_red.clear();
	m_keep.clear();
}
